//
// Project : Jam
// File    : BinaryOperator.java
// Author  : Giovanni Lagorio (gio@libertyline.com)
//

package jam ;

class BinaryOperator extends Expr {

	// arithmetic-logic operators:
	public final static int MULTIPLY 		=  0 ;
	public final static int DIVIDE	 		=  1 ;
	public final static int REMAINDER 		=  2 ;
	public final static int ADD	 		=  3 ;
	public final static int SUBTRACT 		=  4 ;
	public final static int LEFT_SHIFT 		=  5 ;
	public final static int S_RIGHT_SHIFT 		=  6 ;
	public final static int U_RIGHT_SHIFT 		=  7 ;
	public final static int AND	 		=  8 ;
	public final static int EXCLUSIVE_OR 		=  9 ;
	public final static int INCLUSIVE_OR 		= 10 ;
	public final static int CONDITIONAL_AND		= 11 ;
	public final static int CONDITIONAL_OR 		= 12 ;
	// assignments (handled in class Assignment):
	public final static int ASSIGN			= 13 ;
	public final static int ASSIGN_MULTIPLY		= 14 ;
	public final static int ASSIGN_DIVIDE		= 15 ;
	public final static int ASSIGN_REMAINDER	= 16 ;
	public final static int ASSIGN_ADD		= 17 ;
	public final static int ASSIGN_SUBTRACT		= 18 ;
	public final static int ASSIGN_LEFT		= 19 ;
	public final static int ASSIGN_SIGNED_RIGHT	= 20 ;
	public final static int ASSIGN_UNSIGNED_RIGHT	= 21 ;
	public final static int ASSIGN_AND		= 22 ;
	public final static int ASSIGN_EXCLUSIVE_OR	= 23 ;
	public final static int ASSIGN_INCLUSIVE_OR	= 24 ;
	// relational operators (handled in class RelationalOperators):
	public final static int LESS	 		= 25 ;
	public final static int GREATER	 		= 26 ;
	public final static int LESS_EQUAL 		= 27 ;
	public final static int GREATER_EQUAL 		= 28 ;
	public final static int INSTANCE_OF 		= 29 ; // handled in InstanceOfExpr
	public final static int EQUAL	 		= 30 ;
	public final static int NOT_EQUAL 		= 31 ;
	// array access (handled in ArrayAccess)
	public final static int ARRAY_ACCESS		= 32 ;

	static final String [] operators = {
		// arithmetic-logic operators:
		"*", "/", "%", "+", "-", "<<", ">>", ">>>", "&",
		"^", "|", "&&", "||",
		// assignments (handled in class Assignment):
		"=", "*=", "/=", "%=", "+=", "-=", "<<=", ">>=", ">>=", "&=", "^=", "|=",
		// relational operators (handled in class RelationalOperators):
		"<", ">", "<=", ">=", "*** instanceof ***", "==", "!=",
		// array access
		"*** array-access ***" // unused, ArrayAccess overrides toString()
	} ;

	protected Expr e1, e2 ;
	protected int op ;

	public String toString() {
		return e1+operators[op]+e2 ;
	}

	BinaryOperator(int op_, Expr e1_, Expr e2_) {
		if (op_<0 || op_>ARRAY_ACCESS)
			throw new RuntimeException("Internal compiler error (BO op="+op_+")") ;
		op = op_ ;
		e1 = e1_ ;
		e2 = e2_ ;
	}

	void resolveNames(CompilationUnit cu) throws JamException {
		e1.resolveNames(cu) ;
		e2.resolveNames(cu) ;
	}

	Expr typeCheck(Env env, ModuleType currentModule, CompilationUnit cu) throws JamException {
		e1 = e1.typeCheck(env, currentModule, cu) ;
		e2 = e2.typeCheck(env, currentModule, cu) ;
		Type t1 = e1.getType() ;
		Type t2 = e2.getType() ;
		switch (op) {
		case ADD:
			if (t1.toString().equals("java.lang.String")) {
				type = t1 ;
				return this ;
			} else	if (t2.toString().equals("java.lang.String")) {
				type = t2 ;
				return this ;
			}
			// no break here!
		case MULTIPLY:
		case DIVIDE:
		case REMAINDER:
		case SUBTRACT:
			type = Type.binaryNumericPromotion(t1, t2) ;
			return this ;
		case LEFT_SHIFT:
		case S_RIGHT_SHIFT:
		case U_RIGHT_SHIFT:
			type = Type.unaryNumericPromotion(t1) ;
			return this ;
		case AND:
		case EXCLUSIVE_OR:
		case INCLUSIVE_OR:
			if ( t1==PrimitiveType.booleanInstance || t2==PrimitiveType.booleanInstance ) {
				type = PrimitiveType.booleanInstance ;
				return this ;
			}
			type = Type.binaryNumericPromotion(t1, t2) ;
			return this ;
		case CONDITIONAL_AND:
		case CONDITIONAL_OR:
			type = PrimitiveType.booleanInstance ;
			return this ;
		}
		throw new RuntimeException("Internal compiler error (BinaryOperator.typeCheck op="+op+").") ;
	}

}


