//
// Project : Jam
// File    : ClassType.java
// Author  : Giovanni Lagorio (gio@libertyline.com)
//

package jam ;

import java.util.* ;

class ClassType extends ModuleType {

	private ClassType superclass ;
	private MixinType supermixin ;
	private InterfaceType [] interfaces ;

	ClassType(Name name_) throws JamException {
		super(name_) ;
	}

	void init (String superclassName, Vector ifacesV, Constructor [] constructors_, Field [] fields_, Method [] methods_) throws JamException {
		init(superclassName, ifacesV, constructors_, fields_, methods_, null) ;
	}

	void init (String superclassName, Vector ifacesV, Constructor [] constructors_, Field [] fields_, Method [] methods_, MixinType supermixin_) throws JamException {
		constructors = constructors_ ;
		fields = fields_ ;
		methods = methods_ ;
		Type t = Type.getType(superclassName) ;
		if (!(t instanceof ClassType)) 
			throw new JamException("bad 'extends' declaration; \""+t+"\" isn't a class.");
		superclass = (ClassType)t ;
		//System.out.println("supermixin="+supermixin+", superclass="+superclass) ;
		supermixin = supermixin_ ;
		if (supermixin!=null && superclass.isSubtypeOf(supermixin) && fields.length>0)
			System.err.println("Warning: mixin "+supermixin.getModuleName()+" has been instantiated twice on "+superclass.getModuleName()+" ... see the documentation about field resolution through "+supermixin.getModuleName()) ;
		interfaces = new InterfaceType[ifacesV.size()] ;
		for(int a=0; a<interfaces.length; ++a) {
			String i = ((Name)ifacesV.elementAt(a)).toString() ;
			t = Type.getType(i) ;
			if (!(t instanceof InterfaceType))
				throw new JamException("bad 'implements' declaration; \""+i+"\""+
								" isn't an interface.") ;
			interfaces[a] = (InterfaceType)t ;
		}
	}

	ClassType(Class c) throws JamException {
		super(c) ;
		Class sc = c.getSuperclass() ;
		superclass = sc==null ? null : (ClassType)Type.getType(sc) ;
		java.lang.Class [] ifaces = c.getInterfaces() ;
		interfaces = new InterfaceType[ifaces.length] ;
		for(int a=0; a<ifaces.length; ++a)
			interfaces[a] = (InterfaceType)Type.getType(ifaces[a]) ;
		java.lang.reflect.Constructor [] cConstructors = c.getDeclaredConstructors() ;
		constructors = new Constructor[cConstructors.length] ;
		for(int a=0; a<cConstructors.length; ++a) {
			java.lang.reflect.Constructor cc = cConstructors[a] ;
			Modifiers modifiers = new Modifiers(cc.getModifiers()) ;
			String identifier = cc.getName() ;
			Class [] cArgs = cc.getParameterTypes() ;
			Type [] args = new Type[cArgs.length] ;
			for(int j=0; j<args.length; ++j)
				args[j] = getType(cArgs[j]) ;
			Class [] cExcs = cc.getExceptionTypes() ;
			Name [] excs = new Name[cExcs.length] ;
			for(int j=0; j<excs.length; ++j)
				excs[j] = new Name(cExcs[j].getName()) ;
			constructors[a] = new Constructor(this, modifiers, identifier, args, excs) ;
		}
	}

	protected boolean doubleDispatch(Type that) {
		return that.isSubtypeOf(this) ;
	}

	boolean providesMethod(String mname, Type [] paramsType, Type retType, Type [] excs____TO_DO, Modifiers mod) {
		outer: for(int a=0; a<methods.length; ++a) {
			Method m = methods[a] ;
			if ( !m.getIdentifier().equals(mname) ) continue ;
			Type [] args = m.getArguments() ;
			if ( args.length != paramsType.length ) continue ;
			for(int p=0; p<args.length; ++p)
				if (!args[p].equals(paramsType[p])) continue outer ;
			if (! m.getReturnType().equals(retType) )
				return false ;
			Modifiers mm = m.getModifiers() ;
			if ( mm.isSet(Modifiers.ABSTRACT) )
				return mod.isSet(Modifiers.ABSTRACT) ;
			if ( mm.isSet(Modifiers.STATIC) != mod.isSet(Modifiers.STATIC) )
				return false ;
			// 2 do : ... check exceptions ...
			return true ;
		}
		if (superclass==null) return false ;
		return superclass.providesMethod(mname, paramsType, retType, null, mod) ;
	}

	boolean providesField(String fname, Type ftype, Modifiers fmod) throws JamException {
		for(int a=0; a<fields.length; ++a)
			if (fields[a].getIdentifier().equals(fname)) {
				if ( fmod.isSet(Modifiers.STATIC)!=fields[a].getModifiers().isSet(Modifiers.STATIC) )
					return false ;
				return ftype.equals( fields[a].getType() ) ;
			}
		if (superclass==null) return false ;
		return superclass.providesField(fname, ftype, fmod) ;
	}

	Field getField(String identifier) throws JamException {
		for(int a=0; a<fields.length; ++a)
			if (fields[a].getIdentifier().equals(identifier))
				return fields[a] ;
		Field found = getSuperField(identifier) ;
		for(int a=0; a<interfaces.length; ++a) {
			Field f = interfaces[a].getField(identifier) ;
			if (f!=null) {
				if (found!=null)
					throw new JamException("identifier \""+identifier+"\" is ambiguous");
				found = f ;
			}
		}
		return found ;
	}

	Field getSuperField(String identifier) throws JamException {
		if (superclass==null) return null ;
		return superclass.getField(identifier) ;
	}

	Vector getSuperApplicableMethods(String methodName, Type [] args) throws JamException {
		return superclass.getApplicableMethods(methodName, args) ;
	}

	Vector getApplicableMethods(String methodName, Type [] args) throws JamException {
		//System.out.println("ClassType.getApplicableMethods(methodName="+methodName+", args="+args+", args.l="+args.length+")") ;
		//for(int a=0; a<args.length; ++a)
		//	System.out.println("args["+a+"]="+args[a]) ;
		Vector appMethods = superclass==null ? new Vector() :
				superclass.getApplicableMethods(methodName, args) ;
		for(int a=0; a<interfaces.length; ++a) {
			Vector iAppMethods = interfaces[a].getApplicableMethods(methodName, args) ;
			for(Enumeration e=iAppMethods.elements(); e.hasMoreElements();)
				appMethods.addElement(e.nextElement()) ;
		}
		for(int a=0; a<methods.length; ++a) {
			Method m = methods[a] ;
			//System.out.println("m="+m) ;
			//System.out.println("m.getIdentifier().equals(methodName)="+m.getIdentifier().equals(methodName)) ;
			//System.out.println("subtype(args, m.getArguments())="+subtype(args, m.getArguments())) ;
			if ( m.getIdentifier().equals(methodName) && subtype(args, m.getArguments()) )
				appMethods.addElement(m) ;
		}
		return appMethods ;
	}

	boolean isSubtypeOf(MixinType mt) {
		//System.out.println("mt="+mt+", this="+this+", supermixin="+supermixin) ;
		if (supermixin!=null && mt.equals(supermixin)) return true ;
		if (superclass!=null) return superclass.isSubtypeOf(mt) ;
		return false ;
	}

	boolean isSubtypeOf(InterfaceType it) {
		for(int a=0; a<interfaces.length; ++a)
			if (interfaces[a].isSubtypeOf(it)) return true ;
		if (superclass!=null)
			return superclass.isSubtypeOf(it) ;
		return false ;
	}

	boolean isSubtypeOf(ClassType ct) {
		if (toString()==ct.toString()) return true ;
		if (superclass!=null)
			return superclass.isSubtypeOf(ct) ;
		return false ;
	}

	void checkCycles() throws JamException {
		Hashtable visited = new Hashtable() ;
		for( ClassType c = superclass; c!=null; c = c.superclass) {
			if (c==this) 
				throw new JamException("class \""+moduleName+"\": cyclic 'extends' relation.");
			if (visited.get(c)!=null) return ;
			visited.put(c,c) ;
		}
	}

}

