//
// Project : Jam
// File    : Jam.java
// Author  : Giovanni Lagorio (gio@libertyline.com)
//

package jam ;

import java.io.* ;
import java.util.* ;
import invisible.jacc.parse.* ;

public class Jam extends CompilerModel implements PrescannerJavaSourceClient {

	
	private int notInComment;	// The condition number for "notInComment".
	private int inComment;		// The condition number for "inComment".

	private static Name currentPackage ;
	private static CompilationUnit currentCompilationUnit ;
	private static Vector declaredTypes = new Vector() ;
	private static Vector compilationUnits = new Vector() ;
	private static boolean printStackTraces ;
	private static Jam compiler ;

	public Jam(boolean debug) {

		_scannerTable = new JamGrammarScannerTable ();

		notInComment = _scannerTable.lookupCondition ("notInComment");
		inComment = _scannerTable.lookupCondition ("inComment");

		TokenFactory reportCopyOfToken = new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				Token t = new Token(token) ;
				t.column = Token.noPosition ;
				token.value = t ;
				return assemble;
			}} ;
		_scannerTable.linkFactory ("import", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("class", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("interface", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("mixin", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("{", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("}", "", reportCopyOfToken) ;
		_scannerTable.linkFactory ("(", "", reportCopyOfToken) ;
		_scannerTable.linkFactory (")", "", reportCopyOfToken) ;
		_scannerTable.linkFactory (";", "", reportCopyOfToken) ;

		_scannerTable.linkFactory ("lineEnd", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				scanner.countLine ();
				return discard;
			}}) ;

		_scannerTable.linkFactory ("identifier", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				token.value = scanner.tokenToString ();
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("integerLiteral", "illegal", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				reportError (token, null, "Illegal number.");
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("characterLiteral", "illegal", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				reportError (token, null, "Illegal character literal.");
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("stringLiteral", "illegal", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				reportError (token, null, "Illegal string (invalid escape seq.).");
				return assemble;
			}}) ;
		
		_scannerTable.linkFactory ("stringLiteral", "runOn", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				reportError (token, null, "Run-on string.");
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("beginComment", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				scanner.setCondition (inComment);
				return discard;
			}}) ;

		_scannerTable.linkFactory ("endComment", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				scanner.setCondition (notInComment);
				return discard;
			}}) ;

		_scannerTable.linkFactory ("illegalChar", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				scanner.client().scannerUnmatchedToken (scanner, token);
				return discard;
			}}) ;

		TokenFactory reservedKeywordTF = new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				reportError (token, null, "Reserved keyword '" +
							scanner.tokenToString () + "'." );
				return discard ;
			}} ;
		_scannerTable.linkFactory ("const", "", reservedKeywordTF);
		_scannerTable.linkFactory ("goto", "", reservedKeywordTF);

		_scannerTable.linkFactory ("nullLiteral", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
				token.value = new NullLiteral() ;
				return assemble;
			}}) ;
					
		_scannerTable.linkFactory ("booleanLiteral", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new BooleanLiteral(scanner.tokenToString());
				return assemble;
			}}) ;
				
		TokenFactory intLiteralTF = new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new IntLiteral(scanner.tokenToString());
				return assemble;
			}} ;
		_scannerTable.linkFactory ("integerLiteral", "decimal",
								intLiteralTF) ;
		_scannerTable.linkFactory ("integerLiteral", "hex",
								intLiteralTF) ;
		_scannerTable.linkFactory ("integerLiteral", "octal",
								intLiteralTF) ;

		TokenFactory longLiteralTF = new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new LongLiteral(scanner.tokenToString());
				return assemble;
			}} ;
		_scannerTable.linkFactory ("integerLiteral", "decimalLong",
								longLiteralTF) ;
		_scannerTable.linkFactory ("integerLiteral", "hexLong",
								longLiteralTF) ;
		_scannerTable.linkFactory ("integerLiteral", "octalLong",
								longLiteralTF) ;

		_scannerTable.linkFactory ("floatingPointLiteral", "float", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new FloatLiteral(scanner.tokenToString());
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("floatingPointLiteral", "double", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new DoubleLiteral(scanner.tokenToString());
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("characterLiteral", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new CharLiteral(scanner.tokenToString());
				return assemble;
			}}) ;

		_scannerTable.linkFactory ("stringLiteral", "", new TokenFactory() {
			public int makeToken (Scanner scanner, Token token) {
			token.value = new StringLiteral(scanner.tokenToString());
				return assemble;
			}}) ;

		_parserTable = new JamGrammarParserTable ();

		_parserTable.linkFactory ("PrimitiveType", "boolean", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.booleanInstance ;
			}}) ;

		_parserTable.linkFactory ("IntegralType", "byte", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.byteInstance ;
			}}) ;

		_parserTable.linkFactory ("IntegralType", "short", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.shortInstance ;
			}}) ;

		_parserTable.linkFactory ("IntegralType", "int", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.intInstance ;
			}}) ;

		_parserTable.linkFactory ("IntegralType", "long", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.longInstance ;
			}}) ;

		_parserTable.linkFactory ("IntegralType", "char", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.charInstance ;
			}}) ;

		_parserTable.linkFactory ("FloatingPointType", "float", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.floatInstance ;
			}}) ;

		_parserTable.linkFactory ("FloatingPointType", "double", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return PrimitiveType.doubleInstance ;
			}}) ;

		_parserTable.linkFactory ("VoidType", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return VoidType.instance ;
			}}) ;

		_parserTable.linkFactory ("ArrayType", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Object t = parser.rhsValue(0) ;
				Integer d = (Integer)parser.rhsValue(1) ;
				if (t==null || d==null) return null ;
				if (t instanceof Name) t = new TypeName((Name)t) ;
				return ((Type)t).addDimensions( d.intValue() ) ;
			}}) ;

		_parserTable.linkFactory ("DimsOpt", "none", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Integer(0) ;
			}}) ;

		_parserTable.linkFactory ("Dims", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Integer(1) ;
			}}) ;

		_parserTable.linkFactory ("Dims", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Integer i = (Integer)parser.rhsValue(0) ;
				if (i==null) return null ;
				return new Integer(i.intValue()+1) ;
			}}) ;

		_parserTable.linkFactory ("SimpleName", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String id = (String)parser.rhsValue(0) ;
				if (id==null) return null ;
				return new Name(id) ;
			}}) ;

		_parserTable.linkFactory ("QualifiedName", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name n = (Name)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(2) ;
				if (n==null || id==null) return null ;
				n.appendComponent(id) ;
				return n ;
			}}) ;

		_parserTable.linkFactory ("CompilationUnit", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name packName = (Name)parser.rhsValue(0) ;
				Vector imports = (Vector)parser.rhsValue(1) ;
				Vector types = (Vector)parser.rhsValue(2) ;
				if (packName==null || imports==null || types==null)
					return null ;
				currentCompilationUnit = new CompilationUnit(packName, imports, types) ;
				compilationUnits.addElement(currentCompilationUnit) ;
				return null ;
			}}) ;

		_parserTable.linkFactory ("PackageDeclarationOpt", "none", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				currentPackage = new Name() ;
				return currentPackage ;
			}}) ;

		NonterminalFactory returnEmptyVector = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Vector() ;
			}} ;

		_parserTable.linkFactory ("ImportDeclarationsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("TypeDeclarationsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("PackageDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				currentPackage = (Name) parser.rhsValue(1) ;
				return currentPackage ;
			}}) ;

		_parserTable.linkFactory ("ImportDeclarations", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				ImportDeclaration d = (ImportDeclaration)parser.rhsValue(0) ;
				if (d!=null) v.addElement(d) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ImportDeclarations", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				ImportDeclaration decl = (ImportDeclaration)parser.rhsValue(1) ;
				if (decl!=null) v.addElement(decl) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("SingleTypeImportDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name name = (Name)parser.rhsValue(1) ;
				if (name==null) return null ;
				return new ImportDeclaration(name, false, (Token)parser.rhsValue(0)) ;
			}}) ;

		_parserTable.linkFactory ("TypeImportOnDemandDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name name = (Name)parser.rhsValue(1) ;
				if (name==null) return null ;
				return new ImportDeclaration(name, true, (Token)parser.rhsValue(0)) ;
			}}) ;

		_parserTable.linkFactory ("TypeDeclaration", "empty", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return null ;
			}}) ;

		_parserTable.linkFactory ("TypeDeclarations", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				TypeDeclaration td = (TypeDeclaration)parser.rhsValue(0) ;
				if (td!=null) v.addElement(td) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("TypeDeclarations", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				TypeDeclaration td = (TypeDeclaration)parser.rhsValue(1) ;
				if (td!=null) v.addElement(td) ;
				return v ;
			}}) ;
			
		_parserTable.linkFactory ("Modifier", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Modifiers(param) ;
			}}) ;

		_parserTable.linkFactory ("ModifiersOpt", "none", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Modifiers(0) ;
			}}) ;

		_parserTable.linkFactory ("Modifiers", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m1 = (Modifiers)parser.rhsValue(0) ;
				Modifiers m2 = (Modifiers)parser.rhsValue(1) ;
				return m1.addFlags(m2.flags()) ;
			}}) ;

		linkAssignment("assign", Assignment.ASSIGN) ;
		linkAssignment("multiply", Assignment.ASSIGN_MULTIPLY) ;
		linkAssignment("divide", Assignment.ASSIGN_DIVIDE) ;
		linkAssignment("remainder", Assignment.ASSIGN_REMAINDER) ;
		linkAssignment("add", Assignment.ASSIGN_ADD) ;
		linkAssignment("subtract", Assignment.ASSIGN_SUBTRACT) ;
		linkAssignment("left", Assignment.ASSIGN_LEFT) ;
		linkAssignment("signedRight", Assignment.ASSIGN_SIGNED_RIGHT) ;
		linkAssignment("unsignedRight", Assignment.ASSIGN_UNSIGNED_RIGHT) ;
		linkAssignment("and", Assignment.ASSIGN_AND) ;
		linkAssignment("exclusiveOr", Assignment.ASSIGN_EXCLUSIVE_OR) ;
		linkAssignment("inclusiveOr", Assignment.ASSIGN_INCLUSIVE_OR) ;

		_parserTable.linkFactory ("ConditionalExpression", "conditional", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr cond = (Expr)parser.rhsValue(0) ;
				Expr e1 = (Expr)parser.rhsValue(2) ;
				Expr e2 = (Expr)parser.rhsValue(4) ;
				if (cond==null || e1==null || e2==null) return null ;
				return new ConditionalExpression(cond, e1, e2) ;
			}}) ;

		linkBinOp("MultiplicativeExpression", "multiply", BinaryOperator.MULTIPLY) ;
		linkBinOp("MultiplicativeExpression", "divide", BinaryOperator.DIVIDE) ;
		linkBinOp("MultiplicativeExpression", "remainder", BinaryOperator.REMAINDER) ;
		linkBinOp("AdditiveExpression", "add", BinaryOperator.ADD) ;
		linkBinOp("AdditiveExpression", "subtract", BinaryOperator.SUBTRACT) ;
		linkBinOp("ShiftExpression", "left", BinaryOperator.LEFT_SHIFT) ;
		linkBinOp("ShiftExpression", "signedRight", BinaryOperator.S_RIGHT_SHIFT) ;
		linkBinOp("ShiftExpression", "unsignedRight", BinaryOperator.U_RIGHT_SHIFT) ;
		linkBinOp("AndExpression", "and", BinaryOperator.AND) ;
		linkBinOp("ExclusiveOrExpression", "exclusiveOr", BinaryOperator.EXCLUSIVE_OR) ;
		linkBinOp("InclusiveOrExpression", "inclusiveOr", BinaryOperator.INCLUSIVE_OR) ;
		linkBinOp("ConditionalAndExpression", "conditionalAnd", BinaryOperator.CONDITIONAL_AND) ;
		linkBinOp("ConditionalOrExpression", "conditionalOr", BinaryOperator.CONDITIONAL_OR) ;

		linkRelational("RelationalExpression", "less", BinaryOperator.LESS) ;
		linkRelational("RelationalExpression", "greater", BinaryOperator.GREATER) ;
		linkRelational("RelationalExpression", "lessEqual", BinaryOperator.LESS_EQUAL) ;
		linkRelational("RelationalExpression", "greaterEqual", BinaryOperator.GREATER_EQUAL) ;
		linkRelational("RelationalExpression", "instanceof", BinaryOperator.INSTANCE_OF) ;
		_parserTable.linkFactory ("RelationalExpression", "instanceof", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e1 = (Expr)parser.rhsValue(0) ;
				Type t = (Type)parser.rhsValue(2) ;
				if (e1==null || t==null) return null ;
				return new InstanceOfExpr(e1, t) ;
			}}) ;
		linkRelational("EqualityExpression", "equal", BinaryOperator.EQUAL) ;
		linkRelational("EqualityExpression", "notEqual", BinaryOperator.NOT_EQUAL) ;

		linkUnaryOp("UnaryExpressionNotPlusMinus", "not", UnaryOperator.NOT, true) ;
		linkUnaryOp("UnaryExpressionNotPlusMinus", "complement", UnaryOperator.COMPLEMENT, true) ;
		linkUnaryOp("UnaryExpression", "plus", UnaryOperator.PLUS, true) ;
		linkUnaryOp("UnaryExpression", "minus", UnaryOperator.MINUS, true) ;
		linkUnaryOp("PreIncrementExpression", "", UnaryOperator.INCREMENT, true) ;
		linkUnaryOp("PreDecrementExpression", "", UnaryOperator.DECREMENT, true) ;
		linkUnaryOp("PostIncrementExpression", "", UnaryOperator.INCREMENT, false) ;
		linkUnaryOp("PostDecrementExpression", "", UnaryOperator.DECREMENT, false) ;

		_parserTable.linkFactory ("CastExpression", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Object dest = parser.rhsValue(1) ;
				Expr expr = (Expr)parser.rhsValue(3) ;
				if (dest==null || expr==null) return null ;
				if (dest instanceof Type)
					return new CastExpression((Type)dest, expr) ;
				if (dest instanceof Name)
					return new CastExpression(new TypeName((Name)dest), expr) ;
				if (dest instanceof NameExpr)
					return new CastExpression(new TypeName(((NameExpr)dest).getName()), expr) ;
				Token t = (Token)parser.rhsValue(0) ;
				if (t!=null)
					reportError (t, null, "Bad cast expression." );
				else	reportError (null, "Bad cast expression." );
				return null ;
			}}) ;

		_parserTable.linkFactory ("ArrayAccess", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Object o1 = parser.rhsValue(0) ;
				if (o1!=null && o1 instanceof Name)
					o1 = new NameExpr((Name)o1) ;
				Expr e1 = (Expr)o1 ;
				Expr e2 = (Expr)parser.rhsValue(2) ;
				if (e1==null || e2==null) return null ;
				return new ArrayAccess(e1, e2) ;
			}}) ;

		NonterminalFactory nameToExprFactory = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name n = (Name)parser.rhsValue(0) ;
				if (n==null) return null ;
				return new NameExpr(n) ;
			}} ;
		_parserTable.linkFactory ("PostfixExpression", "name", nameToExprFactory) ;
		_parserTable.linkFactory ("LeftHandSide", "name", nameToExprFactory) ;

		_parserTable.linkFactory ("MethodInvocation", "name", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name n = (Name)parser.rhsValue(0) ;
				Vector args = (Vector)parser.rhsValue(2) ;
				if (n==null || args==null) return null ;
				return new NameMethodInvocation(n, args) ;
			}}) ;

		_parserTable.linkFactory ("MethodInvocation", "primary", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr p = (Expr)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(2) ;
				Vector args = (Vector)parser.rhsValue(4) ;
				if (p==null || id==null || args==null) return null ;
				return new PrimaryMethodInvocation(p, id, args) ;
			}}) ;

		_parserTable.linkFactory ("MethodInvocation", "super", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String id = (String)parser.rhsValue(2) ;
				Vector args = (Vector)parser.rhsValue(4) ;
				if (id==null || args==null) return null ;
				return new SuperMethodInvocation(id, args) ;
			}}) ;

		_parserTable.linkFactory ("FieldAccess", "primary", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Object o = parser.rhsValue(0) ;
				if (o!=null && o instanceof Name) o = new NameExpr((Name)o) ;
				Expr p = (Expr)o ;
				String id = (String)parser.rhsValue(2) ;
				if (p==null || id==null) return null ;
				return new PrimaryFieldAccess(p, id) ;
			}}) ;

		_parserTable.linkFactory ("FieldAccess", "super", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String id = (String)parser.rhsValue(2) ;
				if (id==null) return null ;
				return new SuperFieldAccess(id) ;
			}}) ;

		_parserTable.linkFactory ("DimExpr", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Expr)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("DimExprs", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Expr e = (Expr)parser.rhsValue(0) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("DimExprs", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(1) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ArrayCreationExpression", "primitive", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				PrimitiveType t = (PrimitiveType)parser.rhsValue(1) ;
				Vector de = (Vector)parser.rhsValue(2) ;
				Integer dopt = (Integer)parser.rhsValue(3) ;
				if (t==null || de==null || dopt==null) return null ;
				return new PrimitiveArrayCreation(t, de, dopt.intValue())  ;
			}}) ;

		_parserTable.linkFactory ("ArrayCreationExpression", "reference", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name t = (Name)parser.rhsValue(1) ;
				Vector de = (Vector)parser.rhsValue(2) ;
				Integer dopt = (Integer)parser.rhsValue(3) ;
				if (t==null || de==null || dopt==null) return null ;
				return new ReferenceArrayCreation(t, de, dopt.intValue())  ;
			}}) ;

		_parserTable.linkFactory ("ArgumentList", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Object o = parser.rhsValue(0) ;
				Expr e = (o instanceof Name) ?
					new NameExpr((Name)o) :
					(Expr)parser.rhsValue(0) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ArgumentList", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(2) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ArgumentListOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("ClassInstanceCreationExpression", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name cname = (Name)parser.rhsValue(1) ;
				Vector args = (Vector)parser.rhsValue(3) ;
				if (cname==null || args==null) return null ;
				return new ClassCreationExpression(cname, args) ;
			}}) ;

		_parserTable.linkFactory ("PrimaryNoNewArray", "this", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Name("this") ;
			}}) ;

		_parserTable.linkFactory ("PrimaryNoNewArray", "expression", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(1) ;
				if (e==null) return null ;
				return new ParExpr(e) ;
			}}) ;

		_parserTable.linkFactory ("Finally", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Block)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("CatchClause", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				FormalParameter fp = (FormalParameter)parser.rhsValue(2) ;
				Block b = (Block)parser.rhsValue(4) ;
				if (fp==null || b==null) return null ;
				return new CatchClause(fp, b, (Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("Catches", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				CatchClause cc = (CatchClause)parser.rhsValue(0) ;
				if (cc!=null) v.addElement(cc) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("Catches", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				CatchClause cc = (CatchClause)parser.rhsValue(1) ;
				if (cc!=null) v.addElement(cc) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("TryStatement", "catch", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Block b = (Block)parser.rhsValue(1) ;
				Vector catches = (Vector)parser.rhsValue(2) ;
				if (b==null || catches==null) return null ;
				return new TryStatement(b, catches, null) ;
			}}) ;

		_parserTable.linkFactory ("TryStatement", "finally", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Block b = (Block)parser.rhsValue(1) ;
				Block fb = (Block)parser.rhsValue(2) ;
				if (b==null || fb==null) return null ;
				return new TryStatement(b, new Vector(), fb) ;
			}}) ;

		_parserTable.linkFactory ("TryStatement", "catchFinally", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Block b = (Block)parser.rhsValue(1) ;
				Vector catches = (Vector)parser.rhsValue(2) ;
				Block fb = (Block)parser.rhsValue(3) ;
				if (b==null || catches==null || fb==null) return null ;
				return new TryStatement(b, catches, fb) ;
			}}) ;

		_parserTable.linkFactory ("SynchronizedStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(2) ;
				Block b = (Block)parser.rhsValue(4) ;
				if (e==null || b==null) return null ;
				return new SynchronizedStatement(e, b, (Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("ThrowStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(1) ;
				if (e==null) ;
				return new ThrowStatement(e, (Token)parser.rhsValue(2)) ;
			}}) ;

		_parserTable.linkFactory ("ReturnStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new ReturnStatement((Expr)parser.rhsValue(1),
								(Token)parser.rhsValue(2)) ;
			}}) ;

		_parserTable.linkFactory ("BreakStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new BreakContinueStatement( BreakContinueStatement.BREAK,
								(String)parser.rhsValue(1),
								(Token)parser.rhsValue(2)) ;
			}}) ;

		_parserTable.linkFactory ("ContinueStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new BreakContinueStatement( BreakContinueStatement.CONTINUE,
								(String)parser.rhsValue(1),
								(Token)parser.rhsValue(2)) ;
			}}) ;

		_parserTable.linkFactory ("StatementExpressionList", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Expr e = (Expr)parser.rhsValue(0) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("StatementExpressionList", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(2) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		NonterminalFactory forFactory = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Object init = parser.rhsValue(2) ;
				Expr guard = (Expr)parser.rhsValue(4) ;
				Vector update = (Vector)parser.rhsValue(6) ;
				Statement s = (Statement)parser.rhsValue(8) ;
				return new ForStatement(init, guard, update, s, (Token)parser.rhsValue(0)) ;
			}} ;
		_parserTable.linkFactory ("ForStatementNoShortIf", "", forFactory) ;
		_parserTable.linkFactory ("ForStatement", "", forFactory) ;

		_parserTable.linkFactory ("DoStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Statement s = (Statement)parser.rhsValue(1) ;
				Expr e = (Expr)parser.rhsValue(4) ;
				if (s==null || e==null) return null ;
				return new DoStatement(e, s, (Token)parser.rhsValue(5)) ;
			}}) ;

		NonterminalFactory whileFactory = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(2) ;
				Statement s = (Statement)parser.rhsValue(4) ;
				if (e==null || s==null) return null ;
				return new WhileStatement(e, s, (Token)parser.rhsValue(1)) ;
			}} ;
		_parserTable.linkFactory ("WhileStatement", "", whileFactory) ;
		_parserTable.linkFactory ("WhileStatementNoShortIf", "", whileFactory) ;

		_parserTable.linkFactory ("Block", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(1) ;
				if (v==null) return null ;
				return new Block(v) ;
			}}) ;

		_parserTable.linkFactory ("BlockStatementsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("BlockStatements", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Statement s = (Statement)parser.rhsValue(0) ;
				if (s!=null) v.addElement(s) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("BlockStatements", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Statement s = (Statement)parser.rhsValue(1) ;
				if (s!=null) v.addElement(s) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("EmptyStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return EmptyStatement.instance ;
			}}) ;

		NonterminalFactory labeledStatementFactory = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String label = (String)parser.rhsValue(0) ;
				Statement s = (Statement)parser.rhsValue(2) ;
				if (label==null || s==null) return null ;
				return new LabeledStatement(label, s) ;
			}} ;
		_parserTable.linkFactory ("LabeledStatement", "", labeledStatementFactory) ;
		_parserTable.linkFactory ("LabeledStatementNoShortIf", "", labeledStatementFactory) ;

		_parserTable.linkFactory ("ExpressionStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(0) ;
				Token token = (Token)parser.rhsValue(1) ;
				if (e==null || token==null) return null ;
				return new ExpressionStatement(e, token) ;
			}}) ;

		_parserTable.linkFactory ("LocalVariableDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Type t = (Type)parser.rhsValue(0) ;
				Vector d = (Vector)parser.rhsValue(1) ;
				if (t==null || d==null) return null ;
				return new LocalVarDeclaration(t, d) ;
			}}) ;


		_parserTable.linkFactory ("LocalVariableDeclarationStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				LocalVarDeclaration d = (LocalVarDeclaration)parser.rhsValue(0) ;
				if (d==null) return null ;
				return new LocalVarDeclStatement(d, (Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("IfThenStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(2) ;
				Statement ts = (Statement)parser.rhsValue(4) ;
				if (e==null || ts==null) return null ;
				return new IfStatement(e, ts, null, (Token)parser.rhsValue(0)) ;
			}}) ;

		NonterminalFactory ifThenElseFactory = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(2) ;
				Statement ts = (Statement)parser.rhsValue(4) ;
				Statement es = (Statement)parser.rhsValue(4) ;
				if (e==null || ts==null || es==null) return null ;
				return new IfStatement(e, ts, es, (Token)parser.rhsValue(0)) ; 
			}} ;
		_parserTable.linkFactory ("IfThenElseStatement", "", ifThenElseFactory) ;
		_parserTable.linkFactory ("IfThenElseStatementNoShortIf", "", ifThenElseFactory) ;

		_parserTable.linkFactory ("SwitchLabel", "case", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(1) ;
				if (e==null) return null ;
				return new SwitchLabel(e) ;
			}}) ;

		_parserTable.linkFactory ("SwitchLabel", "default", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new SwitchLabel(null) ;
			}}) ;

		_parserTable.linkFactory ("SwitchLabelsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("SwitchLabels", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				SwitchLabel sl = (SwitchLabel)parser.rhsValue(0) ;
				if (sl!=null) v.addElement(sl) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("SwitchLabels", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				SwitchLabel sl = (SwitchLabel)parser.rhsValue(1) ;
				if (sl!=null) v.addElement(sl) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("SwitchBlockStatementGroups", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Vector [/* 2 */] labelsAndStatements = (Vector [])parser.rhsValue(0) ;
				if (labelsAndStatements!=null) v.addElement(labelsAndStatements) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("SwitchBlockStatementGroups", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Vector [/* 2 */] labelsAndStatements = (Vector [])parser.rhsValue(1) ;
				if (labelsAndStatements!=null) v.addElement(labelsAndStatements) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("SwitchBlockStatementGroup", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v1 = (Vector)parser.rhsValue(0) ;
				Vector v2 = (Vector)parser.rhsValue(1) ;
				if (v1==null || v2==null) return null ;
				return new Vector [] {v1, v2} ;
			}}) ;

		_parserTable.linkFactory ("SwitchBlock", "groups", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v1 = (Vector)parser.rhsValue(1) ;
				Vector v2 = (Vector)parser.rhsValue(2) ;
				if (v1==null || v2==null) return null ;
				return new Vector [] {v1, v2} ;
			}}) ;

		_parserTable.linkFactory ("SwitchBlock", "noGroups", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v2 = (Vector)parser.rhsValue(1) ;
				if (v2==null) return null ;
				return new Vector [] {new Vector(), v2} ;
			}}) ;

		_parserTable.linkFactory ("SwitchStatement", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(2) ;
				Vector [] v = (Vector [])parser.rhsValue(4) ;
				if (e==null || v==null) return null ;
				return new SwitchStatement(e, v[0], v[1], (Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("ArrayInitializer", "empty", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new ArrayInitializer(new Vector()) ;
			}}) ;

		_parserTable.linkFactory ("ArrayInitializer", "notEmpty", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(1) ;
				if (v==null) return null ;
				return new ArrayInitializer(v) ;
			}}) ;

		_parserTable.linkFactory ("VariableInitializers", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Expr e = (Expr)parser.rhsValue(0) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("VariableInitializers", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(2) ;
				if (e!=null) v.addElement(e) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("MethodHeader", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers modifiers = (Modifiers)parser.rhsValue(0) ;
				Type type = (Type)parser.rhsValue(1) ;
				String identifier = (String)parser.rhsValue(2) ;
				Token token = (Token)parser.rhsValue(3) ;
				Vector formalParameters = (Vector)parser.rhsValue(4) ;
				Integer dimsOpt = (Integer)parser.rhsValue(6) ;
				Vector exceptions = (Vector)parser.rhsValue(7) ;
				if (modifiers==null || type==null || identifier==null ||
							formalParameters==null || dimsOpt==null
							|| exceptions==null || token==null)
					return null ;
				type = type.addDimensions(dimsOpt.intValue()) ;
				return new MethodHeader(modifiers, type, identifier, 
						formalParameters, exceptions, token) ;
			}}) ;

		_parserTable.linkFactory ("ThrowsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("MethodDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				MethodHeader h = (MethodHeader)parser.rhsValue(0) ;
				Block b = (Block)parser.rhsValue(1) ;
				if (h==null) return null ;
				return new MethodDeclaration(h, b) ;
			}}) ;

		_parserTable.linkFactory ("AbstractMethodDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				MethodHeader h = (MethodHeader)parser.rhsValue(0) ;
				if (h==null) return null ;
				return new MethodDeclaration(h, null) ;
			}}) ;

		_parserTable.linkFactory ("MethodBody", "empty", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return null ; 
			}}) ;

		NonterminalFactory declListFirst = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Decl d = (Decl)parser.rhsValue(0) ;
				if (d!=null) v.addElement(d) ;
				return v ; 
			}} ;
		_parserTable.linkFactory ("InterfaceMemberDeclarations", "first", declListFirst) ;
		_parserTable.linkFactory ("ClassBodyDeclarations", "first", declListFirst) ;
		_parserTable.linkFactory ("MixinBodyDeclarations", "first", declListFirst) ;
		_parserTable.linkFactory ("MixinConstructorDeclarations", "first", declListFirst) ;

		NonterminalFactory declListNext = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Decl d = (Decl)parser.rhsValue(1) ;
				if (d!=null) v.addElement(d) ;
				return v ; 
			}} ;
		_parserTable.linkFactory ("InterfaceMemberDeclarations", "next", declListNext) ;
		_parserTable.linkFactory ("ClassBodyDeclarations", "next", declListNext) ;
		_parserTable.linkFactory ("MixinBodyDeclarations", "next", declListNext) ;
		_parserTable.linkFactory ("MixinConstructorDeclarations", "next", declListNext) ;

		_parserTable.linkFactory ("InterfaceMemberDeclarationsOpt", "none", returnEmptyVector) ;
		_parserTable.linkFactory ("ClassBodyDeclarationsOpt", "none", returnEmptyVector) ;
		_parserTable.linkFactory ("MixinBodyDeclarationsOpt", "none", returnEmptyVector) ;
		_parserTable.linkFactory ("MixinConstructorDeclarationsOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("ExtendsInterfaces", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Name n = (Name)parser.rhsValue(1) ;
				if (n!=null) v.addElement(n) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ExtendsInterfaces", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Name n = (Name)parser.rhsValue(2) ;
				if (n!=null) v.addElement(n) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("ExtendsInterfacesOpt", "none", returnEmptyVector) ;
		_parserTable.linkFactory ("InterfacesOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("InterfaceDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m = (Modifiers)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(2) ;
				Vector ifaces = (Vector)parser.rhsValue(3) ;
				Vector decls = (Vector)parser.rhsValue(5) ;
				if (m==null || id==null || ifaces==null || decls==null) return null ;
				declaredTypes.addElement(currentPackage.copy().appendComponent(id)) ;
				return new InterfaceDeclaration(m, id, ifaces, decls,
								(Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("ExplicitConstructorInvocation", "this", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector args = (Vector)parser.rhsValue(2) ;
				if (args==null) return null ;
				return new ConstructorInvocation(ConstructorInvocation.THIS, args,
								(Token)parser.rhsValue(4)) ;
			}}) ;

		_parserTable.linkFactory ("ExplicitConstructorInvocation", "super", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector args = (Vector)parser.rhsValue(2) ;
				if (args==null) return null ;
				return new ConstructorInvocation(ConstructorInvocation.SUPER, args,
								(Token)parser.rhsValue(4)) ;
			}}) ;

		_parserTable.linkFactory ("ConstructorBody", "explicit", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				ConstructorInvocation ci = (ConstructorInvocation)parser.rhsValue(1) ;
				Vector stmts = (Vector)parser.rhsValue(2) ;
				if (ci==null || stmts==null) return null ;
				return new Object [] {ci, stmts} ;
			}}) ;

		_parserTable.linkFactory ("ConstructorBody", "implicit", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector stmts = (Vector)parser.rhsValue(1) ;
				if (stmts==null) return null ;
				return new Object [] { new ConstructorInvocation(
									ConstructorInvocation.SUPER,
									new Vector(),
									(Token)parser.rhsValue(0)),
									stmts } ;
			}}) ;

		_parserTable.linkFactory ("ConstructorDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m = (Modifiers)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(1) ;
				Vector params = (Vector)parser.rhsValue(3) ;
				Vector exceptions = (Vector)parser.rhsValue(5) ;
				Object [] o = (Object [])parser.rhsValue(6) ;
				ConstructorInvocation ci = (ConstructorInvocation)o[0] ;
				Vector stmts = (Vector)o[1] ;
				if (m==null || id==null || params==null || exceptions==null
								|| ci==null || stmts==null)
					return null ;
				stmts.insertElementAt(ci, 0) ;
				return new ConstructorDeclaration(m, id, params, exceptions, stmts,
								(Token)parser.rhsValue(2)) ;
			}}) ;

		_parserTable.linkFactory ("StaticInitializer", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Block b = (Block)parser.rhsValue(1) ;
				if (b==null) return null ;
				return new StaticInitializer(b) ;
			}}) ;

		NonterminalFactory nameListFirst = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				Name n = (Name)parser.rhsValue(0) ;
				if (n!=null) v.addElement(n) ;
				return v ;
			}} ;
		_parserTable.linkFactory ("ClassTypeList", "first", nameListFirst) ;
		_parserTable.linkFactory ("InterfaceTypeList", "first", nameListFirst) ;

		NonterminalFactory nameListNext = new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				Name n = (Name)parser.rhsValue(2) ;
				if (n!=null) v.addElement(n) ;
				return v ;
			}} ;
		_parserTable.linkFactory ("ClassTypeList", "next", nameListNext) ;
		_parserTable.linkFactory ("InterfaceTypeList", "next", nameListNext) ;

		_parserTable.linkFactory ("Throws", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Vector)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("VariableDeclaratorId", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String id = (String)parser.rhsValue(0) ;
				Integer d = (Integer)parser.rhsValue(1) ;
				if (id==null || d==null) return null ;
				return new VariableDeclaratorId(id, d.intValue()) ;
			}}) ;

		_parserTable.linkFactory ("FormalParameter", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Type t = (Type)parser.rhsValue(0) ;
				VariableDeclaratorId v = (VariableDeclaratorId)parser.rhsValue(1) ;
				if (t==null || v==null) return null ;
				return new FormalParameter(t, v) ;
			}}) ;

		_parserTable.linkFactory ("FormalParameterList", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				FormalParameter fp = (FormalParameter)parser.rhsValue(0) ;
				if (fp!=null) v.addElement(fp) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("FormalParameterList", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				FormalParameter fp = (FormalParameter)parser.rhsValue(2) ;
				if (fp!=null) v.addElement(fp) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("FormalParameterListOpt", "none", returnEmptyVector) ;

		_parserTable.linkFactory ("VariableDeclarator", "noInit", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				VariableDeclaratorId v = (VariableDeclaratorId)parser.rhsValue(0) ;
				if (v==null) return null ;
				return new VariableDeclarator(v, null) ;
			}}) ;

		_parserTable.linkFactory ("VariableDeclarator", "init", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				VariableDeclaratorId v = (VariableDeclaratorId)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(2) ;
				if (v==null || e==null) return null ;
				return new VariableDeclarator(v, e) ;
			}}) ;

		_parserTable.linkFactory ("VariableDeclarators", "first", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = new Vector() ;
				VariableDeclarator vd = (VariableDeclarator)parser.rhsValue(0) ;
				if (vd!=null) v.addElement(vd) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("VariableDeclarators", "next", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Vector v = (Vector)parser.rhsValue(0) ;
				VariableDeclarator vd = (VariableDeclarator)parser.rhsValue(2) ;
				if (vd!=null) v.addElement(vd) ;
				return v ;
			}}) ;

		_parserTable.linkFactory ("FieldDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m = (Modifiers)parser.rhsValue(0) ;
				Type t = (Type)parser.rhsValue(1) ;
				Vector d = (Vector)parser.rhsValue(2) ;
				Token token = (Token)parser.rhsValue(3) ;
				return new FieldDeclaration(m, t, d, token) ;
			}}) ;

		_parserTable.linkFactory ("ClassBody", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Vector)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("MixinBody", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Vector)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("Interfaces", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Vector)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("Super", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Name)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("SuperOpt", "none", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return new Name("java.lang.Object") ;
			}}) ;

		_parserTable.linkFactory ("MixinDeclaration", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				String identifier = (String)parser.rhsValue(1) ;
				Vector ifaces = (Vector)parser.rhsValue(2) ;
				Vector decls = (Vector)parser.rhsValue(3) ;
				if (identifier==null || ifaces==null || decls==null) return null ;
				declaredTypes.addElement(currentPackage.copy().appendComponent(identifier)) ;
				return new MixinDeclaration(identifier, decls, ifaces, (Token)parser.rhsValue(0)) ;
			}}) ;

		_parserTable.linkFactory ("MixinConstructors", "", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				return (Vector)parser.rhsValue(1) ;
			}}) ;

		_parserTable.linkFactory ("ClassDeclaration", "java", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m = (Modifiers)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(2) ;
				Name sclass = (Name)parser.rhsValue(3) ;
				Vector i = (Vector)parser.rhsValue(4) ;
				Vector d = (Vector)parser.rhsValue(5) ;
				if (m==null || id==null || sclass==null || i==null || d==null)
					return null ;
				declaredTypes.addElement(currentPackage.copy().appendComponent(id)) ;
				return new ClassDeclaration(m, id, sclass, i, d,
								(Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("ClassDeclaration", "jam", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Modifiers m = (Modifiers)parser.rhsValue(0) ;
				String id = (String)parser.rhsValue(2) ;
				Name mixin = (Name)parser.rhsValue(4) ;
				Name sclass = (Name)parser.rhsValue(6) ;
				Vector c = (Vector)parser.rhsValue(7) ;
				if (m==null || id==null || mixin==null || sclass==null || c==null)
					return null ;
				declaredTypes.addElement(currentPackage.copy().appendComponent(id)) ;
				return new JamClassDeclaration(m, id, sclass, mixin, c,
								(Token)parser.rhsValue(1)) ;
			}}) ;

		_parserTable.linkFactory ("ReferenceType", "class", new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Name n = (Name)parser.rhsValue(0) ;
				return new TypeName(n) ;
			}}) ;

		if (debug && setDebugMode (true, true))
			throw new InternalError("Compiler: Consistency check failed.");

	}

	private void linkRelational(final String nonterminal,final String link,final int op) {
		_parserTable.linkFactory (nonterminal, link, new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e1 = (Expr)parser.rhsValue(0) ;
				Expr e2 = (Expr)parser.rhsValue(2) ;
				if (e1==null || e2==null) return null ;
				return new RelationalOperator(op, e1, e2) ;
			}}) ;
	}

	private void linkBinOp(final String nonterminal, final String link, final int op) {
		_parserTable.linkFactory (nonterminal, link, new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e1 = (Expr)parser.rhsValue(0) ;
				Expr e2 = (Expr)parser.rhsValue(2) ;
				if (e1==null || e2==null) return null ;
				return new BinaryOperator(op, e1, e2) ;
			}}) ;
	}

	private void linkUnaryOp(final String nonterminal, final String link, final int op,
								final boolean isPrefixed) {
		_parserTable.linkFactory (nonterminal, link, new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr e = (Expr)parser.rhsValue(isPrefixed ? 1:0) ;
				if (e==null) return null ;
				return new UnaryOperator(op, isPrefixed, e) ;
			}}) ;
	}

	private void linkAssignment(final String link,final int op) {
		_parserTable.linkFactory ("Assignment", link, new NonterminalFactory() {
			public Object makeNonterminal (Parser parser, int param) {
				Expr lv = (Expr)parser.rhsValue(0) ;
				Expr e = (Expr)parser.rhsValue(2) ;
				if (lv==null || e==null) return null ;
				return new Assignment(op, lv, e) ;
			}}) ;
	}

	public void javaSourceInvalidEscape (Token token) {
		reportError (token, null, "Invalid escape character." );
	}

	public Scanner makeScanner (String filename) {
		try	{
			InputStream stream = new FileInputStream (filename);
			Prescanner scannerSource =
				new PrescannerJavaSource (this, stream, 2000);
			return Scanner.makeScanner (
					this, scannerSource, _scannerTable,
					filename, 1, 1, 4000, null );
		} catch (IOException e) {
			return null;
		}
	}

	public void scannerEOF (Scanner scanner, Token token) {
		if (scanner.condition() == inComment)
			reportError (token, null, "Run-on comment." );
	}

	private static int parseCommandLineFlags(String [] args) {
		int i = 0 ;
		if ( i<args.length && args[i].equals("-stackTraces") ) {
			printStackTraces = true ;
			++i ;
		}
		if ( i<args.length && args[i].equals("-debug") ) {
			compiler = new Jam(true) ;
			++i ;
		} else	compiler = new Jam(false) ;
		if ( args.length==i ) {
			System.err.println("\nJam 1.05 written by Giovanni Lagorio\n") ;
			System.err.println("Check http://www.disi.unige.it/person/LagorioG/jam for the latest version.") ;
			System.err.println("") ;
			compiler.reportError(null, "no filenames given.") ;
			return -1 ;
		}
		return i ;
	}

	private static boolean parseSources(int startIndex, String args[]) {
		boolean errors = false ;
		for (int i = startIndex; i < args.length; ++i) {
			currentCompilationUnit = null ;
			if (!args[i].toLowerCase().endsWith(".jam")) {
				compiler.reportError(null, args[i]+": source files should have a '.jam' "+
									"extension; skipped") ;
				continue ;
			}
			errors |= compiler.compile (args[i]);
			if (currentCompilationUnit!=null)
				currentCompilationUnit.setFilename(args[i]) ;
		}
		return errors ;
	}

	private static boolean resolveNames() {
		boolean errors = false ;
		for(Enumeration e=compilationUnits.elements(); e.hasMoreElements(); ) {
			CompilationUnit cu = (CompilationUnit)e.nextElement() ;
			try	{
				cu.resolveNames(declaredTypes) ;
			} catch (JamException exc) {
				if (printStackTraces)
					exc.printStackTrace() ;
				Token t = exc.getToken() ;
				String msg = exc.getMessage() ;
				if (t!=null)
					compiler.reportError(t, null, msg) ;
				else	compiler.reportError(null, msg+" (in "+cu.getFilename()+")") ;
				errors = true ;
			}
		}
		return errors ;
	}
	
	private static boolean createModuleObjects() {
		boolean errors = false ;
		for(Enumeration e=compilationUnits.elements(); e.hasMoreElements(); ) { 
			CompilationUnit cu = (CompilationUnit)e.nextElement() ;
			try	{
				cu.createModuleObjects() ;
			} catch (JamException exc) {
				if (printStackTraces)
					exc.printStackTrace() ;
				Token t = exc.getToken() ;
				String msg = exc.getMessage() ;
				if (t!=null)
					compiler.reportError(t, null, msg) ;
				else	compiler.reportError(null, msg+" (in "+cu.getFilename()+")") ;
				errors = true ;
			}
		}
		return errors ;
	}

	private static boolean extractModuleInformation() {
		boolean errors = false ;
		boolean scanMixinDecls = true ;
		for(;;) {
			for(Enumeration e=compilationUnits.elements(); e.hasMoreElements(); ) {
				CompilationUnit cu = (CompilationUnit)e.nextElement() ;
				try	{
					cu.extractModuleInformation(scanMixinDecls) ;
				} catch (JamException exc) {
					if (printStackTraces)
						exc.printStackTrace() ;
					Token t = exc.getToken() ;
					String msg = exc.getMessage() ;
					if (t!=null)
						compiler.reportError(t, null, msg) ;
					else	compiler.reportError(null, msg+" (in "+cu.getFilename()+")") ;
					errors = true ;
				}
			}
			if (scanMixinDecls)
				scanMixinDecls = false ;
			else	break ;
		}
		return errors ;
	}

	private static boolean typeChecking() {
		boolean errors = false ;
		for(Enumeration e=compilationUnits.elements(); e.hasMoreElements(); ) {
			CompilationUnit cu = (CompilationUnit)e.nextElement() ;
			try	{
				cu.typeCheck() ;
			} catch (JamException exc) {
				if (printStackTraces)
					exc.printStackTrace() ;
				Token t = exc.getToken() ;
				String msg = exc.getMessage() ;
				if (t!=null)
					compiler.reportError(t, null, msg) ;
				else	compiler.reportError(null, msg+" (in "+cu.getFilename()+")") ;
				errors = true ;
			}
		}
		return errors ;
	}

	private static boolean checkCycles() {
		boolean errors = false ;
		for(Enumeration e=declaredTypes.elements(); e.hasMoreElements(); ) {
			Name name = (Name)e.nextElement() ;
			try	{
				ModuleType mt = (ModuleType)Type.getType(name.toString()) ;
				mt.checkCycles() ;
			} catch (JamException exc) {
				if (printStackTraces)
					exc.printStackTrace() ;
				Token t = exc.getToken() ;
				String msg = exc.getMessage() ;
				if (t!=null)
					compiler.reportError(t, null, msg) ;
				else	compiler.reportError(null, msg) ;
				errors = true ;
			}
		}
		return errors ;
	}
	
	private static boolean writeJavaFiles() {
		boolean errors = false ;
		for(Enumeration e=compilationUnits.elements(); e.hasMoreElements();) {
			CompilationUnit cu = (CompilationUnit)e.nextElement() ;
			String fname = cu.getFilename() ;
			String newFile = fname.substring(0, fname.length()-1)+"va" ;
			try	{
				PrintStream fout = new PrintStream(new FileOutputStream(newFile)) ;
				fout.println("//\n// File automagically created from \""+fname+"\" --- "+
						"DO NOT EDIT !!!\n//\n// (Jam may overwrite this file"+
						" next time you'll run it)\n//\n") ;
				cu.dump(new TabPrintStream(fout)) ;
				fout.close() ;
				System.gc() ;
			} catch (JamException exc) {
				if (printStackTraces)
					exc.printStackTrace() ;
				Token t = exc.getToken() ;
				String msg = exc.getMessage() ;
				if (t!=null)
					compiler.reportError(t, null, msg) ;
				else	compiler.reportError(null, msg) ;
				errors = true ;
			} catch (IOException ioError) {
				compiler.reportError(null, "Cannot open "+newFile+" for writing.") ;
				errors = true ;
			}

		}
		return errors ;
	}

	public static void main (String[] args) throws Exception {
		int indexOfFirstFilename = parseCommandLineFlags(args) ;
		if ( indexOfFirstFilename<0 ) return ;
		if ( parseSources(indexOfFirstFilename, args) ) return ;
		if ( resolveNames() ) return ;
		if ( createModuleObjects() ) return ;
		if ( extractModuleInformation() ) return ;
		if ( checkCycles() ) return ;
		if ( typeChecking() ) return ;
		if ( !writeJavaFiles() )
			System.out.println("All files have been translated successfully.") ;
	}
}


