//
// Project : Jam
// File    : ModuleType.java
// Author  : Giovanni Lagorio (gio@libertyline.com)
//

package jam ;

import java.util.* ;

abstract class ModuleType extends ReferenceType {

	protected Name moduleName ;
	protected Constructor [] constructors ;
	protected Field [] fields ;
	protected Method [] methods ;
	private String packageName ;

	ModuleType(Name moduleName_) throws JamException {
		moduleName = moduleName_ ;
		Type.newType(moduleName.toString(), this) ;
	}

	ModuleType(Class c) throws JamException {
		moduleName = new Name(c.getName()) ;
		Type.newType(moduleName.toString(), this) ;
		// fields:
		java.lang.reflect.Field [] cFields = c.getDeclaredFields() ;
		fields = new Field[cFields.length] ;
		for(int a=0; a<cFields.length; ++a) {
			java.lang.reflect.Field f = cFields[a] ;
			Modifiers modifiers = new Modifiers(f.getModifiers()) ;
			Type type = getType(f.getType()) ;
			String identifier = f.getName() ;
			fields[a] = new Field(this, modifiers, type, identifier) ;
		}
		// methods:
		java.lang.reflect.Method [] cMethods = c.getDeclaredMethods() ;
		methods = new Method[cMethods.length] ;
		for(int a=0; a<cMethods.length; ++a) {
			java.lang.reflect.Method cm = cMethods[a] ;
			Modifiers modifiers = new Modifiers(cm.getModifiers()) ;
			String identifier = cm.getName() ;
			Type retType = getType(cm.getReturnType()) ;
			Class [] cArgs = cm.getParameterTypes() ;
			Type [] args = new Type[cArgs.length] ;
			for(int j=0; j<args.length; ++j)
				args[j] = getType(cArgs[j]) ;
			Class [] cExcs = cm.getExceptionTypes() ;
			Name [] excs = new Name[cExcs.length] ;
			for(int j=0; j<excs.length; ++j)
				excs[j] = new Name(cExcs[j].getName()) ;
			methods[a] = new Method(this, modifiers, identifier, args, excs, retType, null) ;
		}
	}

	public String toString() { return moduleName.toString() ; }

	String getPackageName() {
		if (packageName!=null) return packageName ;
		String mname = moduleName.toString() ;
		int i = mname.lastIndexOf('.') ;
		return packageName = i==-1 ? "" : mname.substring(0, i) ;
	}

	Name getModuleName() {
		return moduleName ;
	}

	abstract Field getField(String identifier) throws JamException ;
	abstract Field getSuperField(String identifier) throws JamException ;

	abstract Vector getApplicableMethods(String methodName, Type [] args) throws JamException ;
	abstract Vector getSuperApplicableMethods(String methodName, Type [] args) throws JamException ;

	Method getMostSpecificMethod(String methodName, Type [] argTypes, ModuleType currModule) throws JamException {
		//System.out.println("getMostSpecificMethod(methodName="+methodName+", argTypes="+argTypes+", currModule="+currModule) ;
		//for(int a=0; a<argTypes.length; ++a)
		//	System.out.println("argTypes["+a+"]="+argTypes[a]) ;
		return mostSpecificMethod(getApplicableMethods(methodName, argTypes), methodName, currModule) ;
	}

	Method getSuperMostSpecificMethod(String methodName, Type [] argTypes, ModuleType currModule)
											throws JamException {
		return mostSpecificMethod(getSuperApplicableMethods(methodName, argTypes), methodName,
											currModule) ; 
	}

	private Method mostSpecificMethod(Vector applicableMethods, String methodName, ModuleType currModule)
											throws JamException {
		int sz = applicableMethods.size() ;
		if (sz==0)
			throw new JamException("No applicable methods named "+methodName+" found.") ;
		boolean [] remove = new boolean[sz] ;
		for( int a=0; a<sz; ++a ) {
			Method meth = (Method)applicableMethods.elementAt(a) ;
			Modifiers mod = meth.getModifiers() ;
			ModuleType declModule = meth.getDeclaringModule() ;
			if ( mod.isSet(Modifiers.PUBLIC) )
				continue ;
			if ( mod.isSet(Modifiers.PRIVATE) && !declModule.equals(currModule) ) {
				remove[a] = true ;
				continue ;
			}
			if ( declModule.getPackageName().equals(currModule.getPackageName()) )
				continue ;
			if ( mod.isSet(Modifiers.PROTECTED) ) {
				if (currModule.isSubtypeOf(declModule)) continue ;
				remove[a] = true ;
				continue ;
			}
			if (!currModule.isSubtypeOf(declModule))
				remove[a] = true ;
		}
		for( int a=0; a<sz; ++a )
			for(int b=0; b<sz; b++) {
				if (a==b) continue ;
				Method m1 = (Method)applicableMethods.elementAt(a) ;
				Method m2 = (Method)applicableMethods.elementAt(b) ;
				if (m1.isMoreSpecific(m2))
					remove[b] = true ;
			}
		for(int a=sz-1; a>=0; --a)
			if (remove[a])
				applicableMethods.removeElementAt(a) ;
		sz = applicableMethods.size() ;
		if (sz==0)
			throw new JamException("Applicable methods are not accessible.") ;
		if ( sz>1 )
			throw new JamException("method \""+methodName+"\", ambiguous call.") ;
		return (Method)applicableMethods.elementAt(0) ;
	}

	static boolean subtype(Type [] args, Type [] ma) throws JamException {
		//System.out.println("in subtype()") ;
		if ( ma.length!=args.length ) return false ;
		for(int j=0; j<ma.length; ++j) {
			args[j] = Type.getType(args[j]) ;
			ma[j] = Type.getType(ma[j]) ;
			//System.out.println("args["+j+"]="+args[j]+" "+args[j].getClass()) ;
			//System.out.println("ma["+j+"]="+ma[j]+" "+ma[j].getClass()) ;
			if (! args[j].isSubtypeOf(ma[j]) )
				return false ;
		}
		return true ;
	}

	abstract void checkCycles() throws JamException ;

	/*
	public boolean equals(Object o) {
		if (o instanceof ModuleType) {
			return moduleName.equals( ((ModuleType)o).moduleName ) ;
		}
		throw new RuntimeException("Compiler internal error in ModuleType.equals; class is "+o.getClass()) ;
	}
	*/

	public int hashCode() {
		return moduleName.hashCode() ;
	}

}

